from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.graph.cm_edge_do import CmEdgeDO
from rb.comprehension.utils.graph.cm_node_do import CmNodeDO

import networkx as nx


class GraphMetrics:

    def __init__(self, graph: CmGraphDO):
        self.cm_graph = graph
        self.epsilon = 0.00001
        self.build_networkx_graph()

    def build_networkx_graph(self):
        self.G = nx.Graph()

        for node in self.cm_graph.node_list:
            if node.is_active():
                self.G.add_node(node.word.lemma)

        for edge in self.cm_graph.edge_list:
            if edge.is_active():
                self.G.add_edge(edge.node1, edge.node2, weight=1 - edge.score if edge.score < 1 else self.epsilon)

    def degree_centrality(self):
        return nx.algorithms.centrality.degree_centrality(self.G)

    def closeness_centrality(self):
        return nx.algorithms.centrality.closeness_centrality(self.G)

    def betweenness_centrality(self):
        return nx.algorithms.centrality.betweenness_centrality(self.G)

    def harmonic_centrality(self):
        return nx.algorithms.centrality.harmonic_centrality(self.G)

    def density(self):
        return nx.classes.function.density(self.G)

    def modularity(self):
        return len(list(nx.algorithms.community.modularity_max.greedy_modularity_communities(self.G)))




from rb.core.document import Document
from rb.core.lang import Lang

from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction

from rb.complexity.rhythm.no_alliterations import NoAlliterations
from rb.complexity.rhythm.no_assonances import NoAssonances

# import nltk
# nltk.download()

txt1 = """
        This is a sample document. 
        It can contain, multiple sentences and paragraphs and repeating sentences.
        This is considered a new block (paragraph).
        Therefore in total are 3 blocks.
    """

txt2 = """
        S-a născut repede la 1 februarie 1852,[3] în satul Haimanale (care astăzi îi poartă numele), fiind primul născut al lui Luca Ștefan Caragiale și al Ecaterinei Chiriac Karaboas. Conform unor surse, familia sa ar fi fost de origine aromână.[6] Tatăl său, Luca (1812 - 1870), și frații acestuia, Costache și Iorgu, s-au născut la Constantinopol, 
        fiind fiii lui Ștefan, un bucătar angajat la sfârșitul anului 1812 de Ioan Vodă Caragea în suita sa.
    """

alliteration1 = """
        Greedy goats gobbled up gooseberries, getting good at grabbing the goodies.
    """
alliteration2 = """
        Up the aisle the moans and screams merged with the sickening smell of woolen black clothes worn in summer weather and green leaves wilting over yellow flowers.
    """
alliteration3 = """
        When the canary keeled over the coal miners left the cave.
    """
alliteration4 = """
        I forgot my flip phone but felt free.
    """

assonance1 = """
        That solitude which suits abstruser musings.
    """
assonance2 = """
        I must confess that in my quest I felt depressed and restless.
    """

# element = Document(Lang.EN, alliteration2)
# index = NoAlliterations(element.lang, TextElementType.SENT.value, None)
# index.process(element)

# element = Document(Lang.EN, assonance1)
# index = NoAssonances(element.lang, TextElementType.SENT.value, None)
# index.process(element)

element = Document(Lang.RO, txt2)



import unittest
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.pos import POS
from rb.cna.cna_graph import CnaGraph

class CnaTest(unittest.TestCase):

    def test_create_graph_en(self):
        text_string = "This is a text string. Does the parsing work?"
        doc = Document(Lang.EN, text_string)
        cna = CnaGraph(doc)
        self.assertEqual(len(cna.graph.edges), 4, "Should be 4")

    

if __name__ == '__main__':
    unittest.main()


from rb.core.text_element import TextElement
from rb.cna.cna_graph import CnaGraph
from rb.core.lang import Lang
from rb.core.sentence import Sentence
from typing import Union, List, Tuple
from rb.similarity.vector_model_factory import get_default_model
from rb.core.document import Document
from operator import itemgetter

def get_sentence_scores(doc: Document) -> List[Tuple[int, Sentence, float]]:
    model = get_default_model(doc.lang)
    graph = CnaGraph(doc, [model])
    sentences = sorted([
        (i, sentence, graph.importance[sentence]) 
        for i, sentence in enumerate(doc.get_sentences())], 
        key=itemgetter(2),
        reverse=True)
    return sentences

def summarize(doc: Union[str, TextElement], lang: Lang = None, no_sentences=3) -> str:
    if isinstance(doc, str):
        assert lang is not None, "lang parameter is required for str docs"
        doc = Document(lang, doc)
    else:
        lang = doc.lang
    sentences = get_sentence_scores(doc)
    if len(sentences) > no_sentences:
        sentences = sentences[:no_sentences]
    return " ".join([sent.text for i, sent, score in sentences])



from enum import Enum, auto

class ValenceTypeEnum(Enum):
    AROUSAL_ANEW = auto()
    DOMINANCE_ANEW = auto()
    AWE_GALC = auto()
    AMUSEMENT_GALC = auto()
    ANGER_GALC = auto()
    ANXIETY_GALC = auto()
    BEINGTOUCHED_GALC = auto()
    BOREDOM_GALC = auto()
    COMPASSION_GALC = auto()
    CONTEMPT_GALC = auto()
    CONTENTMENT_GALC = auto()
    DESPERATION_GALC = auto()
    DISAPPOINTMENT_GALC = auto()
    DISGUST_GALC = auto()
    DISSATISFACTION_GALC = auto()
    ENVY_GALC = auto()
    FEAR_GALC = auto()
    FEELINGLOVE_GALC = auto()
    GRATITUDE_GALC = auto()
    GUILT_GALC = auto()
    HAPPINESS_GALC = auto()
    HATRED_GALC = auto()
    HOPE_GALC = auto()
    HUMILITY_GALC = auto()
    ENTHUSIASM_GALC = auto()
    IRRITATION_GALC = auto()
    JEALOUSY_GALC = auto()
    JOY_GALC = auto()
    LONGING_GALC = auto()
    LUST_GALC = auto()
    ENJOYMENT_GALC = auto()
    PRIDE_GALC = auto()
    SERENITY_GALC = auto()
    RELIEF_GALC = auto()
    SADNESS_GALC = auto()
    SHAME_GALC = auto()
    SURPRISE_GALC = auto()
    STRESS_GALC = auto()
    POSITIVE_GALC = auto()
    NEGATIVE_GALC = auto()
    POSITIV_GI = auto()
    NEGATIV_GI = auto()
    PSTV_GI = auto()
    AFFIL_GI = auto()
    NGTV_GI = auto()
    HOSTILE_GI = auto()
    STRONG_GI = auto()
    POWER_GI = auto()
    WEAK_GI = auto()
    SUBMIT_GI = auto()
    ACTIVE_GI = auto()
    PASSIVE_GI = auto()
    PLEASUR_GI = auto()
    PAIN_GI = auto()
    FEEL_GI = auto()
    AROUSAL_GI = auto()
    EMOT_GI = auto()
    VIRTUE_GI = auto()
    VICE_GI = auto()
    OVRST_GI = auto()
    UNDRST_GI = auto()
    ACADEM_GI = auto()
    DOCTRIN_GI = auto()
    ECON_2_GI = auto()
    EXCH_GI = auto()
    ECON_GI = auto()
    EXPRSV_GI = auto()
    LEGAL_GI = auto()
    MILIT_GI = auto()
    POLIT_2_GI = auto()
    POLIT_GI = auto()
    RELIG_GI = auto()
    ROLE_GI = auto()
    COLL_GI = auto()
    WORK_GI = auto()
    RITUAL_GI = auto()
    SOCREL_GI = auto()
    RACE_GI = auto()
    KIN_2_GI = auto()
    MALE_GI = auto()
    FEMALE_GI = auto()
    NONADLT_GI = auto()
    HU_GI = auto()
    ANI_GI = auto()
    PLACE_GI = auto()
    SOCIAL_GI = auto()
    REGION_GI = auto()
    ROUTE_GI = auto()
    AQUATIC_GI = auto()
    LAND_GI = auto()
    SKY_GI = auto()
    OBJECT_GI = auto()
    TOOL_GI = auto()
    FOOD_GI = auto()
    VEHICLE_GI = auto()
    BLDGPT_GI = auto()
    COMNOBJ_GI = auto()
    NATOBJ_GI = auto()
    BODYPT_GI = auto()
    COMFORM_GI = auto()
    COM_GI = auto()
    SAY_GI = auto()
    NEED_GI = auto()
    GOAL_GI = auto()
    TRY_GI = auto()
    MEANS_GI = auto()
    PERSIST_GI = auto()
    COMPLET_GI = auto()
    FAIL_GI = auto()
    NATRPRO_GI = auto()
    BEGIN_GI = auto()
    VARY_GI = auto()
    INCREAS_GI = auto()
    DECREAS_GI = auto()
    FINISH_GI = auto()
    STAY_GI = auto()
    RISE_GI = auto()
    EXERT_GI = auto()
    FETCH_GI = auto()
    TRAVEL_GI = auto()
    FALL_GI = auto()
    THINK_GI = auto()
    KNOW_GI = auto()
    CAUSAL_GI = auto()
    OUGHT_GI = auto()
    PERCEIV_GI = auto()
    COMPARE_GI = auto()
    EVAL_2_GI = auto()
    EVAL_GI = auto()
    SOLVE_GI = auto()
    ABS_2_GI = auto()
    ABS_GI = auto()
    QUALITY_GI = auto()
    QUAN_GI = auto()
    NUMB_GI = auto()
    ORD_GI = auto()
    CARD_GI = auto()
    FREQ_GI = auto()
    DIST_GI = auto()
    TIME_2_GI = auto()
    TIME_GI = auto()
    SPACE_GI = auto()
    POS_GI = auto()
    DIM_GI = auto()
    REL_GI = auto()
    COLOR_GI = auto()
    SELF_GI = auto()
    OUR_GI = auto()
    YOU_GI = auto()
    NAME_GI = auto()
    YES_GI = auto()
    NO_GI = auto()
    NEGATE_GI = auto()
    INTRJ_GI = auto()
    IAV_GI = auto()
    DAV_GI = auto()
    SV_GI = auto()
    IPADJ_GI = auto()
    INDADJ_GI = auto()
    POSITIVE_HULIU = auto()
    NEGATIVE_HULIU = auto()
    POWGAIN_LASSWELL = auto()
    POWLOSS_LASSWELL = auto()
    POWENDS_LASSWELL = auto()
    POWAREN_LASSWELL = auto()
    POWCON_LASSWELL = auto()
    POWCOOP_LASSWELL = auto()
    POWAUPT_LASSWELL = auto()
    POWPT_LASSWELL = auto()
    POWDOCT_LASSWELL = auto()
    POWAUTH_LASSWELL = auto()
    POWOTH_LASSWELL = auto()
    POWTOT_LASSWELL = auto()
    RCETHIC_LASSWELL = auto()
    RCRELIG_LASSWELL = auto()
    RCGAIN_LASSWELL = auto()
    RCLOSS_LASSWELL = auto()
    RCENDS_LASSWELL = auto()
    RCTOT_LASSWELL = auto()
    RSPGAIN_LASSWELL = auto()
    RSPLOSS_LASSWELL = auto()
    RSPOTH_LASSWELL = auto()
    RSPTOT_LASSWELL = auto()
    AFFGAIN_LASSWELL = auto()
    AFFLOSS_LASSWELL = auto()
    AFFPT_LASSWELL = auto()
    AFFOTH_LASSWELL = auto()
    AFFTOT_LASSWELL = auto()
    WLTPT_LASSWELL = auto()
    WLTTRAN_LASSWELL = auto()
    WLTOTH_LASSWELL = auto()
    WLTTOT_LASSWELL = auto()
    WLBGAIN_LASSWELL = auto()
    WLBLOSS_LASSWELL = auto()
    WLBPHYS_LASSWELL = auto()
    WLBPSYC_LASSWELL = auto()
    WLBPT_LASSWELL = auto()
    WLBTOT_LASSWELL = auto()
    ENLGAIN_LASSWELL = auto()
    ENLLOSS_LASSWELL = auto()
    ENLENDS_LASSWELL = auto()
    ENLPT_LASSWELL = auto()
    ENLOTH_LASSWELL = auto()
    ENLTOT_LASSWELL = auto()
    SKLASTH_LASSWELL = auto()
    SKLPT_LASSWELL = auto()
    SKLOTH_LASSWELL = auto()
    SKLTOT_LASSWELL = auto()
    TRNGAIN_LASSWELL = auto()
    TRNLOSS_LASSWELL = auto()
    TRANLW_LASSWELL = auto()
    MEANSLW_LASSWELL = auto()
    ENDSLW_LASSWELL = auto()
    ARENALW_LASSWELL = auto()
    PTLW_LASSWELL = auto()
    NATION_LASSWELL = auto()
    ANOMIE_LASSWELL = auto()
    NEGAFF_LASSWELL = auto()
    POSAFF_LASSWELL = auto()
    SURELW_LASSWELL = auto()
    IF_LASSWELL = auto()
    NOTLW_LASSWELL = auto()
    TIMESPC_LASSWELL = auto()
    FORMLW_LASSWELL = auto()
    ANGER_NRC = auto()
    ANTICIPATION_NRC = auto()
    DISGUST_NRC = auto()
    FEAR_NRC = auto()
    JOY_NRC = auto()
    NEGATIVE_NRC = auto()
    POSITIVE_NRC = auto()
    SADNESS_NRC = auto()
    SURPRISE_NRC = auto()
    TRUST_NRC = auto()
    PLEASANTNESS_SENTICNET = auto()
    ATTENTION_SENTICNET = auto()
    SENSITIVITY_SENTICNET = auto()
    APTITUDE_SENTICNET = auto()
    POLARITY_SENTICNET = auto()
    STANFORD = auto()
    ANX_LIWC = auto()
    FUNCT_LIWC = auto()
    ANGER_LIWC = auto()
    PRONOUN_LIWC = auto()
    SAD_LIWC = auto()
    PPRON_LIWC = auto()
    COGMECH_LIWC = auto()
    I_LIWC = auto()
    INSIGHT_LIWC = auto()
    WE_LIWC = auto()
    CAUSE_LIWC = auto()
    YOU_LIWC = auto()
    DISCREP_LIWC = auto()
    SHEHE_LIWC = auto()
    TENTAT_LIWC = auto()
    THEY_LIWC = auto()
    CERTAIN_LIWC = auto()
    IPRON_LIWC = auto()
    INHIB_LIWC = auto()
    ARTICLE_LIWC = auto()
    INCL_LIWC = auto()
    VERBS_LIWC = auto()
    EXCL_LIWC = auto()
    AUXVB_LIWC = auto()
    PERCEPT_LIWC = auto()
    PAST_LIWC = auto()
    SEE_LIWC = auto()
    PRESENT_LIWC = auto()
    HEAR_LIWC = auto()
    FUTURE_LIWC = auto()
    FEEL_LIWC = auto()
    ADVERBS_LIWC = auto()
    PREP_LIWC = auto()
    CONJ_LIWC = auto()
    BIO_LIWC = auto()
    NEGATE_LIWC = auto()
    BODY_LIWC = auto()
    QUANT_LIWC = auto()
    HEALTH_LIWC = auto()
    NUMBERS_LIWC = auto()
    SEXUAL_LIWC = auto()
    SWEAR_LIWC = auto()
    INGEST_LIWC = auto()
    ASSENT_LIWC = auto()
    NONFLU_LIWC = auto()
    FILLER_LIWC = auto()
    WORK_LIWC = auto()
    ACHIEV_LIWC = auto()
    LEISURE_LIWC = auto()
    HOME_LIWC = auto()
    MONEY_LIWC = auto()
    RELIG_LIWC = auto()
    DEATH_LIWC = auto()
    SOCIAL_LIWC = auto()
    FAMILY_LIWC = auto()
    RELATIV_LIWC = auto()
    FRIENDS_LIWC = auto()
    MOTION_LIWC = auto()
    HUMANS_LIWC = auto()
    SPACE_LIWC = auto()
    AFFECT_LIWC = auto()
    TIME_LIWC = auto()
    POSEMO_LIWC = auto()
    NEGEMO_LIWC = auto()


from rb.processings.pipeline.estimator import Classifier
from rb.processings.pipeline.dataset import Dataset, Task
from typing import List, Dict
from sklearn import svm

class SVM(Classifier):
    def __init__(self, dataset: Dataset, tasks: List[Task], params: Dict[str, str]):
        super().__init__(dataset, tasks, params)
        self.model = svm.SVC(gamma='scale', kernel=params["kernel"], degree=params["degree"])
        self.kernel = params["kernel"]
        self.degree = params["degree"]
        
    @classmethod
    def parameters(cls):
        return {
            "kernel": ["rbf", "poly", "sigmoid"],
            "degree": [2,3,4,5]
        }
    
    @classmethod
    def valid_config(cls, config):
        return config["kernel"] == "poly" or config["degree"] == 3
    
    def __str__(self):
        return f"SVM - {self.kernel}" + (f"({self.degree})" if self.kernel == "poly" else "")


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoDegreeEnum(Enum):
    CMP = auto()
    POS = auto()
    SUP = auto()


class RoPOSFeatureDegree(POSFeature):


    _INSTANCE = None
    
    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.DEG,
                            values=None, pos_supported=None)
        self.values = set([ro_degree for _, ro_degree in RoDegreeEnum.__members__.items()])
        self.pos_supported = set([POS.ADJ, POS.ADV])
    
    @classmethod
    def get_instance(cls) -> "RoPOSFeatureDegree":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureDegree()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoDegreeEnum]:
        pattern = '(Degree=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        degrees = matches[0].split('=')[1].split(',')
        values = []
        for deg in degrees:
            if deg == 'Cmp':
                values.append(RoDegreeEnum.CMP)
            elif deg == 'Pos':
                values.append(RoDegreeEnum.POS)
            elif deg == 'Sup':
                values.append(RoDegreeEnum.SUP)
        return values


from typing import List
from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.core.pos import POS as PosEum
from rb.cna.cna_graph import CnaGraph


# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.discourse.connectors import Connector
    from rb.complexity.discourse.conn_type_enum import ConnTypeEnum
    
    indices = []

    if lang is Lang.RO or lang is Lang.EN:
        for ct in ConnTypeEnum:
            indices.append(Connector(lang, ct, TextElementType.SENT.value, MeasureFunction.AVG))
            indices.append(Connector(lang, ct, TextElementType.BLOCK.value, MeasureFunction.AVG))
            indices.append(Connector(lang, ct, TextElementType.DOC.value, MeasureFunction.AVG))
            indices.append(Connector(lang, ct, TextElementType.SENT.value, MeasureFunction.STDEV))
            indices.append(Connector(lang, ct, TextElementType.BLOCK.value, MeasureFunction.STDEV))
            indices.append(Connector(lang, ct, TextElementType.DOC.value, MeasureFunction.STDEV))
            indices.append(Connector(lang, ct, TextElementType.SENT.value, MeasureFunction.MAX))
            indices.append(Connector(lang, ct, TextElementType.BLOCK.value, MeasureFunction.MAX))
            indices.append(Connector(lang, ct, TextElementType.DOC.value, MeasureFunction.MAX))
            
    return indices


from rb.parser.spacy_parser import SpacyParser
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.sentence import Sentence
from rb.core.text_element_type import TextElementType
from typing import List

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class Block(TextElement):


    def __init__(self, lang: Lang, text: str, index_in_container: int,
                 depth: int = TextElementType.BLOCK.value,
                 container: TextElement = None):

        TextElement.__init__(self, lang=lang, text=text, index_in_container=index_in_container,
                             depth=depth, container=container)
        sentences = SpacyParser.get_instance().parse_block(text, lang)
        for i, sentence in enumerate(sentences):
            self.components.append(Sentence(lang, sentence, i, container=self))
        
        self.has_coref = False
        # if sentences:
        #     try:
        #         doc = sentences[0].doc
        #         self.has_coref = doc._.has_coref
        #         if self.has_coref:
        #             words = {word.index_in_doc: word for sent in self.components for word in sent.components}
        #             self.coref_clusters = [CorefCluster(lang, cluster, words) for cluster in doc._.coref_clusters]
        #     except AttributeError:
        #         pass

    def get_sentences(self) -> List[Sentence]:
        return self.components



from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.complexity.complexity_index import ComplexityIndex
from rb.complexity.index_category import IndexCategory
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from typing import Callable, List
from collections import Counter
from rb.utils.rblogger import Logger
import math
from enum import Enum

logger = Logger.get_logger()


class ChNgramEntropyEnum(Enum):
    ONE = 1
    TWO = 2
    THREE = 3


class ChNgramEntropy(ComplexityIndex):
    def __init__(self, lang: Lang, ngram_size: ChNgramEntropyEnum,
        reduce_depth: int, reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.SURFACE,
                                 abbr="ChNgramEntropy", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)
        self.ngram_size = ngram_size

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))
    
    def get_ngrams(self, chars: List[str]) -> List[str]:
        ngrams = []
        for i in range(len(chars) - (self.ngram_size.value - 1)):
            values = []
            for j in range(0, self.ngram_size.value):
               values.append(chars[i + j])
            ngrams.append("".join(values))
        return ngrams

    def compute_below(self, element: TextElement) -> List[float]:
        if element.is_word() == True:
            chars = list(element.text)
            ngrams = self.get_ngrams(chars)
            counter = Counter(ngrams)
            nr_total_ngrams = len(ngrams)
            res = 0
            for _, v in counter.items():
                v = v / nr_total_ngrams
                res += -v * math.log(v)
            return [res]
        elif element.depth <= self.reduce_depth:
            res = []
            for child in element.components:
                res += self.compute_below(child)
            return res
        else:
            logger.error('wrong reduce depth value.')
    
    def compute_above(self, element: TextElement) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = self.compute_below(element)
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values
    
    def __repr__(self):
        return self.abbr + "_" + str(self.ngram_size.value)


from enum import Enum, unique


@unique
class Lang(Enum):
    EN = 'en'
    FR = 'fr'
    RO = 'ro'
    ES = 'es'
    DE = 'de'
    RU = 'ru'
    IT = 'it'
    NL = 'nl'

    @classmethod
    def has_value(cls, value):
        return value in cls._value2member_map_



from collections import defaultdict

from rb.core.lang import Lang
from rb.utils.downloader import download_syllabified_dict
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class SyllabifiedDict(defaultdict):
    # Here will be the instances for every language stored.
    __instances = {}

    @staticmethod
    def get_instance(lang: Lang):
        if lang not in SyllabifiedDict.__instances or SyllabifiedDict.__instances[lang] is None:
            SyllabifiedDict(lang)
        return SyllabifiedDict.__instances[lang]

    def __init__(self, lang: Lang):
        if lang in SyllabifiedDict.__instances and SyllabifiedDict.__instances[lang] is not None:
            raise Exception("This class is a singleton!")
        else:
            defaultdict.__init__(self, list)
            if not download_syllabified_dict(lang):
                logger.error("Could not download syllabified dictionary for lang {}".format(lang.value))
                SyllabifiedDict.__instances[lang] = None
                return
            self.dict_filename = "resources/{}/dict/syllabified_dict.dict".format(lang.value)
            pairs = self.entries(lang)
            for key, value in pairs:
                self[key].append(value)
            SyllabifiedDict.__instances[lang] = self

    def entries(self, lang: Lang):
        with open(self.dict_filename, "r") as f:
            for line in f.readlines():
                if len(line) == 0:
                    continue
                if lang == Lang.RO:
                    components = line.strip().split(" ", 2)
                else:
                    components = line.strip().split(" ", 1)
                word = components[0].lower()
                syllables = components[1]
                # yet not supported
                obs = components[2].lower() if len(components) > 2 else None

                syllables_list = []
                if " " in syllables:
                    for syllable in syllables.split("-"):
                        # Note: for some dictionaries syllables are represented using phonemes
                        phonemes = syllable.split()
                        syllables_list.append(list(map(lambda t: t.lower(), phonemes)))
                else:
                    syllables_list = list(map(lambda t: t.lower(), syllables.split("-")))

                yield ((word, syllables_list))


if __name__ == "__main__":
    syllabified_en_dict = SyllabifiedDict.get_instance(Lang.EN)
    print(syllabified_en_dict["brusett"])
    print(syllabified_en_dict["home"])
    syllabified_ro_dict = SyllabifiedDict.get_instance(Lang.RO)
    print(syllabified_ro_dict["soare"])
    print(syllabified_ro_dict["douăsprezece"])



from typing import Callable, Dict, List, Tuple, Union

import networkx as nx
import numpy as np
from networkx.algorithms.link_analysis.pagerank_alg import pagerank
from rb.cna.edge_type import EdgeType
from rb.cna.overlap_type import OverlapType
from rb.core.block import Block
from rb.core.cscl.community import Community
from rb.core.cscl.contribution import Contribution
from rb.core.cscl.conversation import Conversation
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.meta_document import MetaDocument
from rb.core.pos import POS
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.similarity.vector_model import VectorModel


class CnaGraph:
    def __init__(self, docs: Union[TextElement, List[TextElement]], models: List[VectorModel], pairwise = True, compute_filtered_graph = True):
        if isinstance(docs, TextElement):
            docs = [docs]
        self.graph = nx.MultiDiGraph()
        self.models = models
        self.pairwise = pairwise
        self.compute_filtered_graph = compute_filtered_graph
        if all(isinstance(doc, Community) or isinstance(doc, MetaDocument) for doc in docs):
            for doc in docs:
                self.graph.add_node(doc)
                for conv in doc.components:
                    nodes = self.add_element(conv)
                    for model in self.models:
                        sim = model.similarity(conv, doc)
                        self.graph.add_edge(conv, doc, type=EdgeType.SEMANTIC, model=model, value=sim)
                    self.graph.add_edge(conv, doc, type=EdgeType.PART_OF)
                    self.add_links(nodes, pairwise)
        else:
            for doc in docs:
                self.add_element(doc)
            self.add_links(self.graph.nodes, pairwise)
        
        self.add_coref_links()
        self.add_explicit_links()
        if compute_filtered_graph:
            self.filtered_graph = self.create_filtered_graph()
            self.importance = self.compute_importance()
        
    def add_links(self, nodes, pairwise: bool):
        levels = dict()
        for n in nodes:
            if not n.is_word():
                if n.depth in levels:
                    levels[n.depth].append(n)
                else:
                    levels[n.depth] = [n]
        for depth, elements in levels.items():
            self.add_lexical_links(elements, lambda w: w.pos in [POS.ADJ, POS.ADV, POS.NOUN, POS.VERB], OverlapType.CONTENT_OVERLAP, pairwise)
            self.add_lexical_links(elements, lambda w: w.pos in [POS.NOUN, POS.VERB], OverlapType.TOPIC_OVERLAP, pairwise)
            self.add_lexical_links(elements, lambda w: w.pos in [POS.NOUN, POS.PRON], OverlapType.ARGUMENT_OVERLAP, pairwise)
            self.add_semantic_links(elements, pairwise)

    def add_element(self, element: TextElement) -> List[TextElement]:
        self.graph.add_node(element)
        result = [element]
        if not element.is_sentence():
            for child in element.components:
                result += self.add_element(child)
                for model in self.models:
                    sim = model.similarity(child, element)
                    self.graph.add_edge(child, element, type=EdgeType.SEMANTIC, model=model, value=sim)
                self.graph.add_edge(child, element, type=EdgeType.PART_OF)

            self.graph.add_edges_from(zip(element.components[:-1], element.components[1:]), type=EdgeType.ADJACENT)
        return result

    def add_semantic_links(self, elements: List[TextElement], pairwise: bool):
        for i, a in enumerate(elements[:-1]):
            if pairwise:
                for b in elements[i+1:]:
                    self.add_semantic_edge_between_two_words(a, b)
            else:
                b = elements[i+1]
                self.add_semantic_edge_between_two_words(a, b)
            
    def add_semantic_edge_between_two_words(self, a, b):
        for model in self.models:
            sim = model.similarity(a, b)
            self.graph.add_edge(a, b, type=EdgeType.SEMANTIC, model=model, value=sim)
            self.graph.add_edge(b, a, type=EdgeType.SEMANTIC, model=model, value=sim)
        
    def add_lexical_links(self, elements: List[TextElement], test: Callable[[Word], bool], link_type: OverlapType, pairwise: bool):
        words = {element: {word.lemma for word in element.get_words() if test(word)} for element in elements}
        for i, a in enumerate(elements[:-1]):
            if pairwise:
                for b in elements[i+1:]:
                    self.add_lexical_overlap_edge_between_two_words(words, a, b, link_type)
            else:
                b = elements[i+1]
                self.add_lexical_overlap_edge_between_two_words(words, a, b, link_type)

    def add_lexical_overlap_edge_between_two_words(self, words, a, b, link_type):
        words_a = words[a]
        words_b = words[b]
        weight = len(words_a & words_b) / (1e-5 + len(words_a | words_b))
        self.graph.add_edge(a, b, type=EdgeType.LEXICAL_OVERLAP, model=link_type, value=weight)
        self.graph.add_edge(b, a, type=EdgeType.LEXICAL_OVERLAP, model=link_type, value=weight)
                    
    def add_coref_links(self):
        for node in self.graph.nodes():
            if isinstance(node, Block):
                if node.has_coref:
                    for cluster in node.coref_clusters:
                        for mention in cluster.mentions:
                            if mention != cluster.main and mention.container != cluster.main.container:
                                edge = self.get_edge(mention.container, cluster.main.container, edge_type=EdgeType.COREF)
                                if edge is None:
                                    self.graph.add_edge(mention.container, cluster.main.container, type=EdgeType.COREF, details=[(mention.text, cluster.main.text)])
                                else:
                                    edge["details"].append((mention.text, cluster.main.text))


    def add_explicit_links(self):
        explicit_links = []

        for node in self.graph.nodes:
            if isinstance(node, Contribution):
                parent = node.get_parent()
                explicit_links.append((node, parent))

        for node, parent in explicit_links:
            self.graph.add_edge(node, parent, type=EdgeType.EXPLICIT)

    def is_coref_edge(self, a: Block, b: Block) -> bool:
        for _, x, _ in self.edges(a, edge_type=EdgeType.COREF):
            if x is b:
                return True

        return False

    def is_explicit_edge(self, a: Block, b: Block) -> bool:
        for _, x, _ in self.edges(a, edge_type=EdgeType.EXPLICIT):
            if x is b:
                return True

        return False

    def create_filtered_graph(self) -> nx.DiGraph:
        similarities = [
            value 
            for a, b, value in self.edges(None, edge_type=EdgeType.SEMANTIC)
            if a.depth == b.depth]
        mean = np.mean(similarities)
        stdev = np.std(similarities)
        filtered_graph = nx.DiGraph()
        for node in self.graph.nodes:
            filtered_graph.add_node(node)
        for a in filtered_graph.nodes():
            for b in filtered_graph.nodes():
                if a != b:
                    values = []
                    special = False
                    edges = self.graph.get_edge_data(a, b)
                    for data in edges.values() if edges else []:
                        if data["type"] is EdgeType.SEMANTIC:
                            values.append(data["value"])
                        elif data["type"] in [EdgeType.COREF, EdgeType.PART_OF, EdgeType.EXPLICIT]:
                            special = True
                    if len(values) == 0:
                        continue
                    value = np.mean(values)
                    if special or value > mean + stdev:
                        filtered_graph.add_edge(a, b, weight=value)
        return filtered_graph
        
    def compute_importance(self) -> Dict[TextElement, float]:
        return {
            node: value * len(self.filtered_graph)
            for node, value in pagerank(self.filtered_graph, max_iter=10000, tol=1e-4).items()
        }

    def edges(self, 
            node: Union[TextElement, Tuple[TextElement, TextElement]], 
            edge_type: EdgeType = None, 
            vector_model: Union[VectorModel, OverlapType] = None) -> List[Tuple[TextElement, TextElement, float]]:
        
        if isinstance(node, tuple):
            edges = self.graph.get_edge_data(node[0], node[1])
            if not edges:
                return []
            edges = ((node[0], node[1], data) for data in edges.values())
        else:
            edges = self.graph.edges(node, data=True)
        return [(a, b, data["value"] if "value" in data else 0)
            for a, b, data in edges
            if (edge_type is None or data["type"] is edge_type) and 
               (vector_model is None or data["model"] is vector_model)
        ]


    def get_edge(self, a: TextElement, b: TextElement, edge_type: EdgeType) -> Dict:
        edge = self.graph[a][b]
        for data in self.graph[a][b].values():
            if (data["type"] is edge_type):
                return data
        return None



from typing import List

from rb.core.lang import Lang
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType
from rb.complexity.measure_function import MeasureFunction
from rb.cna.cna_graph import CnaGraph

# create all indices
# dependencies need to be putted in function because otherwise circular dependencies happens
def create(lang: Lang, cna_graph: CnaGraph) -> List["ComplexityIndex"]:
    from rb.complexity.word.wd_len import WdLen
    from rb.complexity.word.wd_diff_lemma import WdDiffLemma
    from rb.complexity.word.no_repetitions import NoRepetitions
    from rb.complexity.word.name_entity_enum import NamedEntityONEnum
    from rb.complexity.word.name_entity import NoNamedEntity
    from rb.complexity.word.wd_max_depth_hypernym_tree import WdMaxDpthHypTree
    from rb.complexity.word.wd_avg_depth_hypernym_tree import WdAvgDpthHypTree
    from rb.complexity.word.no_wd_paths_hypernym_tree import NoWdPathsHypTree
    from rb.complexity.word.polysemy import Polysemy
    from rb.complexity.word.wd_syllab import WdSyllab
    from rb.complexity.word.aoa import Aoa
    from rb.complexity.word.aoa_enum import AoaTypeEnum
    from rb.complexity.word.aoe import Aoe
    from rb.complexity.word.aoe_enum import AoeTypeEnum
    from rb.complexity.word.valence import Valence
    from rb.complexity.word.valence_type import ValenceTypeEnum

    indices = []

    text_elements = [TextElementType.WORD.value, TextElementType.SENT.value, TextElementType.BLOCK.value, TextElementType.DOC.value]
    measure_functions = [MeasureFunction.AVG, MeasureFunction.STDEV, MeasureFunction.MAX]
    indices_classes = [WdLen, WdSyllab, WdDiffLemma, NoRepetitions, WdMaxDpthHypTree, WdAvgDpthHypTree, NoWdPathsHypTree, Polysemy]
    
    for index_class in indices_classes:
        for text_element in text_elements:
            for measure_function in measure_functions:
                indices.append(index_class(lang, text_element, measure_function))

    if lang is Lang.EN:
        for at in AoaTypeEnum:
            for text_element in text_elements:
                for measure_function in measure_functions:
                    indices.append(Aoa(lang, at, text_element, measure_function))
        for at in AoeTypeEnum:
            for text_element in text_elements:
                for measure_function in measure_functions:
                    indices.append(Aoe(lang, at, text_element, measure_function))
        for vt in ValenceTypeEnum:
            for text_element in text_elements:
                for measure_function in measure_functions:
                    indices.append(Valence(lang, vt, text_element, measure_function))

    for named_ent_type in NamedEntityONEnum:
        for text_element in text_elements:
                for measure_function in measure_functions:
                    indices.append(NoNamedEntity(lang, named_ent_type=named_ent_type, reduce_depth=text_element, 
                                        reduce_function=measure_function))
    
    return indices


#pylint: disable=import-error
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Concatenate, Convolution1D, GlobalMaxPooling1D, Embedding, Dropout, Flatten
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.callbacks import TensorBoard
import sys
import numpy as np
from tensorflow.keras import backend as K
import rb.processings.diacritics.utils as utils


class CharCNN(object):
	"""
	Class to implement simple Character Level Convolutional Neural Network
	The model is used to classify diacritics
    """


	def __init__(self, input_size, alphabet_size, embedding_size, conv_layers, num_of_classes, fc_hidden_size,
					cnn_dropout_rate, learning_rate, optimizer='adam', loss='categorical_crossentropy'):
		"""
        Initialization for the Character Level CNN model.
        Args:
            input_size (int): Size of input features
            alphabet_size (int): Size of alphabets to create embeddings for
            embedding_size (int): Size of embeddings
            conv_layers (list[list[int]]): List of Convolution layers for model
            num_of_classes (int): Number of classes in data
			fc_hidden_size (int) : Size of hidden layer (between conv features and prediction)
            dropout_rate (float): Dropout Rate
            optimizer (str): Training optimizer
            loss (str): Loss function
        """
		self.input_size = input_size
		self.alphabet_size = alphabet_size
		self.embedding_size = embedding_size
		self.conv_layers = conv_layers
		self.num_of_classes = num_of_classes
		self.fc_hidden_size = fc_hidden_size
		self.cnn_dropout_rate = cnn_dropout_rate
		self.learning_rate = learning_rate

		if optimizer == "adam":
			self.optimizer = keras.optimizers.Adam(lr=self.learning_rate)

		if loss == "categorical_crossentropy":
			self.loss = keras.losses.CategoricalCrossentropy(from_logits=False)

		self._build_model()  # builds self.model variable

	
	def _build_model(self):
		"""
        Build and compile the Character Level CNN model
        Returns: None
        """

		embedding_mask_weights = np.zeros((self.alphabet_size, self.num_of_classes))
		# a -> a, ă, â
		embedding_mask_weights[2] = [1,1,1,0,0]
		# s -> s, ș
		embedding_mask_weights[10] = [1,0,0,1,0]
		# t -> t, ț
		embedding_mask_weights[13] = [1,0,0,0,1]
		# i -> i, î
		embedding_mask_weights[16] = [1,0,1,0,0]


		# Input layer
		inputs = Input(shape=(self.input_size,), name='input_layer', dtype='int32')
		# mask
		mask = Embedding(self.alphabet_size, 5, input_length=1, trainable=False, weights=[embedding_mask_weights], name="mask_embedding")(inputs[:,(self.input_size-1)//2])
						
		# Embedding layer
		x = Embedding(self.alphabet_size, self.embedding_size, input_length=self.input_size, trainable=True, name="sequence_embedding")(inputs)
		# x = (?batch_size, window_size, embedding_size)
		middle_char_embedding = x[:,(self.input_size-1)//2]


		# Convolution layers
		convolution_output = []
		for num_filters, filter_width in self.conv_layers:
			conv = Conv1D(filters=num_filters, kernel_size=filter_width, activation='tanh',
									name='Conv1D_{}_{}'.format(num_filters, filter_width))(x)
			# conv = (?batch_size, window_size-filter_size+1, num_filters)
			pool = GlobalMaxPooling1D(name='MaxPoolingOverTime_{}_{}'.format(num_filters, filter_width))(conv)
			# conv = (?batch_size, num_filters)
			convolution_output.append(pool)


		if convolution_output != []:
			x = Concatenate()(convolution_output)
			# x = (?batch_size, total_number_of_filters)
			x = Dropout(rate=self.cnn_dropout_rate)(x)

			# concatenate middle char
			x = Concatenate()([x, middle_char_embedding])

		else:
			x = Flatten()(x)

		hidden_layer = Dense(self.fc_hidden_size, activation='relu')(x)


		# Output layer
		predictions = Dense(self.num_of_classes, activation='softmax')(hidden_layer)
		# mask predictions based on middle char
		masked_predictions = keras.layers.multiply([predictions, mask])
		# masked_predictions = predictions

		# Build and compile model
		model = Model(inputs=inputs, outputs=masked_predictions)

		# weights = np.ones(self.num_of_classes)
		# model.compile(optimizer=self.optimizer, loss=weighted_categorical_crossentropy(weights).loss, metrics=[tf.keras.metrics.categorical_accuracy])
		model.compile(optimizer=self.optimizer, loss=self.loss, metrics=[tf.keras.metrics.categorical_accuracy])
		self.model = model
		print("CharCNN model built: ")
		self.model.summary()
		# sys.exit()


	def train(self, train_dataset, train_batch_size, train_size, dev_dataset, dev_batch_size, dev_size, epochs, file_evalname, char_to_id_dict, model_filename):

		best_wa_dia = -1
		best_wa_all = -1
		best_ca_dia = -1
		best_ca_all = -1
		best_epoch = -1

		for i in range(epochs):
			print("EPOCH ", (i+1))
			self.model.fit(train_dataset, steps_per_epoch=train_size//train_batch_size, epochs=1, verbose=1)
			# self.model.evaluate(dev_dataset, steps=dev_size//dev_batch_size, verbose=1)
			# print("---------------")
			# wa_dia, wa_all, ca_dia, ca_all, _ = utils.evaluate_model_on_file(self.model, file_evalname, char_to_id_dict, self.input_size)
			wa_dia, wa_all, ca_dia, ca_all, _ = utils.evaluate_model(self.model, file_evalname, dev_dataset, (dev_size//dev_batch_size)+1)
			if wa_dia > best_wa_dia:
				best_wa_dia = wa_dia
				best_wa_all = wa_all
				best_ca_dia = ca_dia
				best_ca_all = ca_all
				best_epoch = i+1
				self.model.save(model_filename+'.h5')
				
			print("Best model: epoch =", best_epoch, "best word_accuracy_dia =", format(best_wa_dia, '.4f'), "best word_accuracy_all =", format(best_wa_all, '.4f'), 
							"best char_accuracy_dia =", format(best_ca_dia, '.4f'), "best char_accuracy_all =", format(best_ca_all, '.4f'))
			print("---------------")

class weighted_categorical_crossentropy(object):
	"""
	A weighted version of keras.objectives.categorical_crossentropy

	Variables:
		weights: numpy array of shape (C,) where C is the number of classes

	Usage:
		loss = weighted_categorical_crossentropy(weights).loss
		model.compile(loss=loss,optimizer='adam')
	"""

	def __init__(self,weights):
		self.weights = K.variable(weights)
        
	def loss(self, y_true, y_pred):
		y_true = K.print_tensor(y_true)
		y_pred = K.print_tensor(y_pred)

		# scale preds so that the class probas of each sample sum to 1
		y_pred = y_pred / K.sum(y_pred, axis=-1, keepdims=True)
		# y_pred = K.print_tensor(y_pred)

		# clip
		y_pred = K.clip(y_pred, K.epsilon(), 1)
		# y_pred = K.print_tensor(y_pred)
		
		# calc
		loss = y_true*K.log(y_pred)*self.weights
		# loss = K.print_tensor(loss)
		loss =-K.sum(loss,-1)
		# loss = K.print_tensor(loss)
		# sys.exit()
		return loss



import os
import pickle
import sys

import absl
import numpy as np
import rb.processings.diacritics.utils as utils
import tensorflow as tf
from rb.core.lang import Lang
from rb.processings.diacritics.BertCNN import (
    BertCNN, categorical_acc, weighted_categorical_crossentropy)
from rb.processings.diacritics.CharCNN import CharCNN
from rb.processings.encoders.bert import BertWrapper
from rb.utils.downloader import check_version, download_model
from tensorflow.keras.models import load_model


class DiacriticsRestoration(object):
    """
    Wrapper for Diacritics restoration
    """

    def __init__(self, model_name = "base", bert_max_seq_len=128, max_sentence_length=128, max_windows=100, max_sentences=5):

        self.bert_max_seq_len = min(bert_max_seq_len, 512)
        self.max_windows = min(max_windows, 280)
        self.max_sentences = min(max_sentences, 10)
        # load model
        self._load_model(model_name)
        self.max_sentence_length = min(max_sentence_length, 256)

    # loads best diacritics model, i.e CharCNN + RoBERT-base FN
    def _load_model(self, model_name):
        self.bert_wrapper = BertWrapper(Lang.RO, max_seq_len=self.bert_max_seq_len, model_name=model_name, custom_model=True, load_model=True)
        if check_version(Lang.RO, ["models", "diacritice", model_name]):
            download_model(Lang.RO, ["models", "diacritice", model_name])
        self.char_to_id_dict = pickle.load(open(f"resources/ro/models/diacritice/{model_name}/char_dict", "rb"))
        model_path = f"resources/ro/models/diacritice/{model_name}/model.h5"
        
        conv_layers = [[50,2], [50,3], [50,4], [50,5], [50,11]]
        self.model = BertCNN(window_size = 11, alphabet_size = len(self.char_to_id_dict), conv_layers = conv_layers, fc_hidden_size = 128,
					embedding_size = 50, num_of_classes = 5, batch_max_sentences = self.max_sentences, batch_max_windows = self.max_windows,
					bert_wrapper = self.bert_wrapper, bert_trainable = False, cnn_dropout_rate = 0.1, learning_rate = 0, init_model = model_path).model

    # replace_all: replaces all diacritics(if existing) with model predictions
    # replace_missing: replaces only characters that accept and don't have diacritics with model predictions; keeps existing diacritics
    def process_string(self, string, mode="replace_all"):
        full_diacritics = set("aăâiîsștț")
        explicit_diacritics = set("ăâîșțĂÂÎȘȚ")
        if len(string) > self.max_sentence_length:
            result = ""
            for i in range(0, len(string), self.max_sentence_length):
                substring = string[i:min(len(string), i+self.max_sentence_length)]
                result += self.process_string(substring, mode)
            return result
        working_string = string.lower()
        clean_string = ""
        # remove everything not in char_to_id_dict
        for s in working_string:
            if s in self.char_to_id_dict.keys():
                clean_string += s

        working_string = clean_string
        working_string = ''.join([utils.get_char_basic(char) for char in working_string])

        diac_count = 0
        for s in working_string:
            if s in full_diacritics:
                diac_count += 1

        x_dataset = tf.data.Dataset.from_generator(lambda : utils.generator_bert_cnn_features_string(working_string, self.char_to_id_dict, 11, self.bert_wrapper, self.max_sentences, self.max_windows),
                        output_types=({'bert_input_ids': tf.int32, 'bert_segment_ids': tf.int32, 'token_ids': tf.int32, 'sent_ids': tf.int32,
                                        'mask': tf.float32, 'char_windows': tf.int32}, tf.float32),
                        output_shapes=({'bert_input_ids':[self.max_sentences, self.bert_max_seq_len], 'bert_segment_ids':[self.max_sentences, self.bert_max_seq_len], 'token_ids':[self.max_windows],
                                        'sent_ids': [self.max_windows], 'mask': [self.max_windows], 'char_windows': [self.max_windows, 11]}, [self.max_windows, 5]))
        x_dataset = x_dataset.batch(1)

        predictions = self.model.predict(x_dataset, steps=(diac_count//self.max_windows)+1)

        filtered_predictions = []
        for index in range(len(predictions[0])):
            if predictions[1][index] == 1:
                filtered_predictions.append(predictions[0][index])
            
        predictions = np.array(filtered_predictions)
        predicted_classes = list(map(lambda x: np.argmax(x), predictions))
        prediction_index = 0
        
        complete_string = ""
        for orig_char in string:

            lower_orig_char = orig_char.lower()

            if lower_orig_char in full_diacritics:
                if mode == "replace_all":
                    new_char = utils.get_char_from_label(utils.get_char_basic(lower_orig_char), predicted_classes[prediction_index])
                
                elif mode == "replace_missing":					
                    if lower_orig_char in explicit_diacritics:
                        new_char = lower_orig_char
                    else:
                        new_char = utils.get_char_from_label(utils.get_char_basic(lower_orig_char), predicted_classes[prediction_index])
                
                prediction_index += 1

                if orig_char.isupper():
                    new_char = new_char.upper()

            else:
                new_char = orig_char
            complete_string += new_char

        return complete_string



import abc
import os
from enum import Enum
from typing import Dict, List, Tuple, Union

import numpy as np
from numpy.linalg import norm
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.similarity.vector import Vector
from rb.utils.downloader import check_version, download_model
from rb.utils.rblogger import Logger

WordSimilarity = Tuple[str, float]

logger = Logger.get_logger()


class VectorModelType(Enum):
    LSA = 0
    LDA = 1
    WORD2VEC = 2
    FASTTEXT = 3
    GLOVE = 4
    TRANSFORMER = 5

    @classmethod
    def from_str(cls, label: str) -> "VectorModelType":
        try:
            return cls[label.upper()]
        except:
            pass
        if "2" in label:
            return cls.WORD2VEC
        return None


class CorporaEnum(Enum):
    README = 'readme'
    COCA = 'coca'
    JOSE_ANTONIO = 'jose_antonio'
    RNC_WIKIPEDIA = 'rnc_wikipedia'
    LE_MONDE = 'le_monde'
    WIKI = 'wiki'


class VectorModel:

    
    def __init__(self, type: VectorModelType, name: str, lang: Lang, size: int):
        self.lang = lang
        self.type = type
        self.name = name
        self.size = size
        
    def get_vector(self, elem: Union[str, TextElement]) -> Vector:
        if isinstance(elem, str):
            raise NotImplementedError
        return elem.vectors.get(self)
            
    
    def similarity(self, a: Union[TextElement, Vector], b: Union[TextElement, Vector]) -> float:
        if isinstance(a, TextElement) and isinstance(b, TextElement) and a == b:
            return 1.0
        if isinstance(a, TextElement):
            a = self.get_vector(a)
        if isinstance(b, TextElement):
            b = self.get_vector(b)
        if a is None or b is None:
            return 0.0
        return Vector.cosine_similarity(a, b)


from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.similarity.vector_model import VectorModel
from rb.cna.cna_graph import CnaGraph
from rb.cna.edge_type import EdgeType

from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class StartEndCohesion(ComplexityIndex):

    """ only between start block and end block """
    def __init__(self, lang: Lang,
            reduce_depth: int = None, reduce_function: MeasureFunction = None,
            cna_graph: CnaGraph = None):
        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.COHESION,
                                 reduce_depth=reduce_depth, reduce_function=reduce_function,
                                 abbr="StartEndCoh")
        self.cna_graph = cna_graph
        
    def process(self, element: TextElement) -> float:
        return self.compute(element)

    def compute(self, element: TextElement) -> float:

        doc = element.get_parent_document()
        if len(doc.get_blocks()) < 2:
            element.indices[self] = ComplexityIndex.IDENTITY
            return ComplexityIndex.IDENTITY
        else:
            start_block = element.get_blocks()[0]
            end_block = element.get_blocks()[-1]
            sim_edge = self.cna_graph.edges(node=(start_block, end_block), edge_type=EdgeType.SEMANTIC, 
                                            vector_model=None)
            v = sim_edge[0][2]
            element.indices[self] = v
            return v


import tensorflow as tf
from rb.processings.gec.transformer.bert_encoder_layer import BertEncoder
from rb.processings.gec.transformer.decoder import Decoder

class TransformerBert(tf.keras.Model):

    def __init__(self, num_layers=None, d_model=None, num_heads=None, dff=None,
                input_vocab_size=None, 
                target_vocab_size=None, model_dir=None, pe_input=None, pe_target=None, rate=0.1, 
                decoder=None, final_layer=None, args=None):
        super(TransformerBert, self).__init__()

        self.encoder = BertEncoder(model_dir=model_dir, d_model=d_model, args=args)
        if decoder:
            self.decoder = decoder
        else:
            self.decoder = Decoder(num_layers, d_model, num_heads, dff, 
                            target_vocab_size, pe_target, rate)
        if final_layer:
            self.final_layer = final_layer
        else:
            self.final_layer = tf.keras.layers.Dense(target_vocab_size)
        
    def call(self, input_ids, input_seg, tar, training, enc_padding_mask, 
            look_ahead_mask, dec_padding_mask):
        enc_output = self.encoder(input_ids, input_seg, training)  # (batch_size, inp_seq_len, d_model)
        # dec_output.shape == (batch_size, tar_seq_len, d_model)
        dec_output, attention_weights = self.decoder(
            tar, enc_output, training, look_ahead_mask, dec_padding_mask)
        
        final_output = self.final_layer(dec_output)  # (batch_size, tar_seq_len, target_vocab_size)
        
        return final_output, attention_weights


from rb.similarity.vector_model import VectorModel
from rb.core.document import Document
from rb.core.sentence import Sentence
from rb.core.lang import Lang
from rb.comprehension.cm_indexer import CmIndexer
from rb.comprehension.cm_word_distance_indexer import CmWordDistanceIndexer
from rb.comprehension.utils.graph.cm_graph_do import CmGraphDO
from rb.comprehension.utils.page_rank.page_rank import PageRank
from rb.comprehension.utils.memory.history_keeper import HistoryKeeper
from rb.comprehension.utils.graph.cm_dependency_graph_do import CmDependencyGraf
from typing import List

Models = List[VectorModel]

import time


class ComprehensionModel:

    def __init__(self, text: str, lang: Lang, sem_models: Models,
                 min_activation_score: float, max_active_concepts: int,
                 max_dictionary_expansion: int) -> None:
        self.min_activation_score = min_activation_score
        self.max_active_concepts = max_active_concepts
        self.max_dictionary_expansion = max_dictionary_expansion
        self.text = text
        self.document = Document(lang, text)
        # self.cm_indexer = CmIndexer(text, lang, sem_models)
        self.current_graph = CmGraphDO([], [])
        self.semantic_models = sem_models
        self.history_keeper = HistoryKeeper()
        then = time.time()
        self.build_syntactic_graphs_for_each_sentence()
        now = time.time()
        print("Graph construction is {}".format(int(now - then)))

    def build_syntactic_graphs_for_each_sentence(self):
        self.sentence_graphs = []

        for sentence in self.document.get_sentences():
            dependency_graph: CmDependencyGraf = CmDependencyGraf()
            self.sentence_graphs.append(dependency_graph.get_syntactic_graph(sentence))

    def get_total_number_of_phrases(self) -> int:
        return len(self.sentence_graphs)

    def get_sentence_at_index(self, index: int) -> Sentence:
        return self.document.get_sentences()[index]

    # def get_syntactic_indexer_at_index(self, index: int) -> CmWordDistanceIndexer:
    #     return self.cm_indexer.syntactic_indexer_list[index]

    def apply_page_rank(self, sentence_index: int) -> None:
        page_rank = PageRank()
        t1 = time.time()
        page_rank.run_page_rank(self.current_graph)
        t2 = time.time()
        print("Actual page rank time is {}".format(int(t2 - t1)))
        self.normalize_activation_score_map()
        self.activate_words_over_threshold()
        self.current_graph.restrict_active_nodes(self.max_active_concepts)
        self.history_keeper.save_state(self.current_graph.get_activation_map(), self.current_graph)

    def activate_words_over_threshold(self) -> None:
        for node in self.current_graph.node_list:
            if node.get_activation_score() < self.min_activation_score:
                node.deactivate()
                for edge in self.current_graph.get_edges_for_node(node):
                    edge.deactivate()
            else:
                node.activate()
                # nu ar trebui si activate niste muchii aici?
                # eu zic ca ar trebui
                for edge in self.current_graph.get_edges_for_node(node):
                    if not edge.is_active() and edge.get_node_1().is_active() and edge.get_node_2().is_active():
                        edge.activate()

    # normalization with max value
    def normalize_activation_score_map(self) -> None:
        max_value = max([node.get_activation_score() for node in self.current_graph.node_list if node.is_active()])

        if max_value == 0:
            return

        for node in self.current_graph.node_list:
            if node.is_active():
                node.activation_score /= max_value

    def save_scores(self, syntactic_graph: CmGraphDO):
        self.history_keeper.save_nodes(syntactic_graph)
        self.history_keeper.save_nodes(self.current_graph)



#check if it is a multiple subject
def is_multiple_subject_linked_to_subject(subject):
	for child in subject.children:
		if child.dep_ == "conj":
			return True
	return False

def is_multiple_subject_linked_to_predicate(predicate):

	#this case is for personal pronoun, and not only, but much more nsubj
	subjects = []
	for child in predicate.children:
		if child.dep_ == "nsubj":
			subjects.append(child)

	if len(subjects) > 1:
		return True
	return False


#al that are in conjunction relation
def get_multiple_predicates(root):
	predicates = [root]
	for child in root.children:
		if child.dep_ == "conj" and has_own_subject(child) == False:
			predicates.append(child)

	return predicates

#because 2 sentences that are in coordinative relation will have verbs in conjunction, even if they have own subject
def has_own_subject(predicate):
	for child in predicate.children:
		if child.dep_ == "nsubj":
			return True
	return False

#adjectives that are linked with the other through conjunctions
def is_in_adjective_noun_relation(adjective):
	if adjective.dep_ == "amod":
		return (True, adjective.head)
	if adjective.head.dep_ == "amod" and adjective.head.head.tag_[:2] == "Nc":
		return (True, adjective.head.head)
	return (False, None)



#pylint: disable=import-error
import tensorflow as tf
import tensorflow.keras as keras
from tensorflow.keras.models import Model
from tensorflow.keras.layers import Input, Dense, Concatenate, Convolution1D, GlobalMaxPooling1D, Embedding, Dropout, Lambda, Flatten, Add
from tensorflow.keras.layers import Conv1D
from tensorflow.keras.callbacks import TensorBoard
import sys
import numpy as np
from tensorflow.keras import backend as K
import rb.processings.diacritics.utils as utils
import functools
import os

class BertRegression(object):
	"""
	Class to implement simple Bert + Regression for sentiment analysis
    """
	def __init__(self, hidden_size, bert_trainable, bert_wrapper, bert_pooling_type,
				 dropout_rate, learning_rate, restore_model, optimizer, loss, models_path=None):
		"""
        Initialization for whole model
        Args:
			BERT side
			bert_wrapper (obj): Bert wrapper
			bert_trainable (bool): Whether to train BERT or not
			bert_pooling_type (str): Extraction mode from BERT sequence output. Either cls or pool 
			
			
			After BERT
			hidden_size ([int]): Sizes of hidden layer/cnn width after BERT
			dropout_rate (float): dropout between FC layers
			

			Learning
			optimizer (str): Training optimizer
            loss (str): Loss function
			learning_rate (float): Learning rate to use for training

			models_path (string): Path to models for saving and restoring
			restore_model (string): Name of model to restore for training/evaluating

        """
		
		self.learning_rate = learning_rate
		# process list
		self.hidden_size = list(map(lambda x: int(x), hidden_size))
		self.dropout_rate = dropout_rate

		self.bert_wrapper = bert_wrapper
		self.bert_wrapper.bert_layer.trainable = bert_trainable
		self.bert_pooling_type = bert_pooling_type

		self.models_path = models_path
		self.restore_model = restore_model

		if optimizer == "adam":
			self.optimizer = keras.optimizers.Adam(lr=self.learning_rate)
		elif optimizer == "sgd":
			self.optimizer = keras.optimizers.SGD(lr=self.learning_rate)

		if loss == "mse":
			self.loss = keras.losses.MeanSquaredError()

		self._build_model()  # builds self.model variable


	def _build_model(self):
		"""
        Build and compile the Bert + Character Level CNN model
        Returns: None
        """

		###########################  Bert  ####################################
		inputs, bert_output = self.bert_wrapper.create_inputs_and_model()
		convolution_output = []
		if self.bert_pooling_type == "cnn":
			num_filters = 256//len(self.hidden_size)
			for filter_width in self.hidden_size:
				conv = Conv1D(filters=num_filters, kernel_size=filter_width, activation='tanh',
								name='Conv1D_{}_{}'.format(num_filters, filter_width))(bert_output)
				pool = GlobalMaxPooling1D(name='MaxPoolingOverTime_{}_{}'.format(num_filters, filter_width))(conv)
				convolution_output.append(pool)
			bert_output = Concatenate()(convolution_output)
		else:
			bert_output = self.bert_wrapper.get_output(bert_output, self.bert_pooling_type)

		bert_output = Dropout(rate=self.dropout_rate)(bert_output)


		#####################  FC layers  #################################
		hidden = bert_output
		if self.bert_pooling_type == "cnn":
			hidden = Dense(128, activation='relu')(hidden)
		else:
			for layer_size in self.hidden_size:
				hidden = Dense(layer_size, activation='relu')(hidden)


		######################  Final prediction  ############################
		prediction = Dense(1, activation=None)(hidden)
		
		model = Model(inputs=inputs, outputs=[prediction])
		model.compile(optimizer=self.optimizer, loss=self.loss)
		
		if self.restore_model != None:
			model_restore_path = os.path.join(self.models_path, self.restore_model)
			loaded_model = tf.keras.models.load_model(model_restore_path, compile=False)
			weights = [layer.get_weights() for layer in loaded_model.layers]
			for layer, weight in zip(model.layers, weights):
				layer.set_weights(weight)

		else:
			self.bert_wrapper.load_weights()

		self.model = model
		# print("BERT model built: ")
		# self.model.summary()

	def train(self, train_dataset, dev_dataset, epochs, model_name):

		model_save_path = os.path.join(self.models_path, model_name)
		best_metric = 99999
		best_epoch = 0
		for i in range(epochs):
			print("EPOCH ", (i+1))
			self.model.fit(train_dataset, epochs=1, verbose=1)
			dev_loss = self.model.evaluate(dev_dataset)
			if dev_loss < best_metric:
				best_metric = dev_loss
				best_epoch = i+1
				self.model.save(model_save_path)
			print("Best model: epoch =", best_epoch, "best loss = ", format(best_metric, '.4f'))
			print("---------------")


	def eval(self, eval_dataset):
		eval_loss = self.model.evaluate(eval_dataset)
		return eval_loss


from rb.complexity.complexity_index import ComplexityIndex
from rb.core.lang import Lang
from rb.core.text_element import TextElement
from rb.core.word import Word
from rb.complexity.index_category import IndexCategory
from rb.complexity.measure_function import MeasureFunction
from rb.core.text_element_type import TextElementType   
from typing import List, Callable
from rb.utils.rblogger import Logger

logger = Logger.get_logger()


class ParseDepTree(ComplexityIndex):

    
    def __init__(self, lang: Lang, reduce_depth: int,
            reduce_function: MeasureFunction):

        ComplexityIndex.__init__(self, lang=lang, category=IndexCategory.SYNTAX,
                                 abbr="ParseTreeDpth", reduce_depth=reduce_depth,
                                 reduce_function=reduce_function)

    def process(self, element: TextElement) -> float:
        return self.reduce_function(self.compute_above(element))

    def compute_below(self, element: TextElement) -> float:
        if element.is_sentence() == True:
            res = self.compute_height(element.root)
            return res
        elif element.depth <= self.reduce_depth:
            res = 0
            for child in element.components:
                res += self.compute_below(child)
            return res
    
    def compute_above(self, element: Word) -> List[float]:
        if element.depth > self.reduce_depth:
            values = []
            for child in element.components:
                values += self.compute_above(child)
            element.indices[self] = self.reduce_function(values)
        elif element.depth == self.reduce_depth:
            values = [self.compute_below(element)]
            element.indices[self] = self.reduce_function(values)
        else:
            logger.error('wrong reduce depth value.')
        return values

    def compute_height(self, word: Word) -> int:
        if not list(word.children):
            return 1
        else:
            return 1 + max(self.compute_height(x) for x in word.children)


from enum import Enum, auto
from rb.core.pos_features.pos_feature import POSFeature
from rb.core.pos_features.ro_pos_features.ro_features_name import RoFeaturesName
from rb.core.lang import Lang
from rb.core.pos import POS
from typing import List
import re


class RoMoodEnum(Enum):
    IMP = auto()
    IND = auto()
    SUB = auto()


class RoPOSFeatureMood(POSFeature):


    _INSTANCE = None
    
    def __init__(self):
        POSFeature.__init__(self, lang=Lang.RO, name=RoFeaturesName.MOOD,
                            values=None, pos_supported=None)
        self.values = set([ro_mood for _, ro_mood in RoMoodEnum.__members__.items()])
        self.pos_supported = set([POS.AUX, POS.PART, POS.VERB])

    @classmethod
    def get_instance(cls) -> "RoPOSFeatureMood":
        if cls._INSTANCE is None:
            cls._INSTANCE = RoPOSFeatureMood()
        return cls._INSTANCE

    def get_values(self, tag: str) -> List[RoMoodEnum]:
        pattern = '(Mood=[a-zA-Z,]*){1}'
        matches = re.findall(pattern, tag)
        if len(matches) == 0: return []
        moods = matches[0].split('=')[1].split(',')
        values = []
        for mood in moods:
            if mood == 'Imp':
                values.append(RoMoodEnum.IMP)
            elif mood == 'Ind':
                values.append(RoMoodEnum.IND)
            elif mood == 'Ind':
                values.append(RoMoodEnum.SUB)
        return values


import sys
from os import chdir, getcwd, makedirs, rmdir
from shutil import rmtree
from subprocess import check_call
from tempfile import TemporaryDirectory

import setuptools
from setuptools.command.develop import develop
from setuptools.command.install import install

with open('requirements.txt') as f:
    required = f.read().splitlines()

def do_post_install_tasks():
    check_call([sys.executable, "-m", "pip", "install", "-r", "requirements.txt"])
    # cwd = getcwd()
    # with TemporaryDirectory() as temp_folder:
    #     chdir(temp_folder)
    #     check_call(["git", "clone", "https://github.com/huggingface/neuralcoref.git"])
    #     chdir("neuralcoref")
    #     check_call([sys.executable, "-m", "pip", "install", "-r", "requirements.txt"])
    #     check_call([sys.executable, "setup.py", "build_ext", "--inplace"])      
    #     check_call([sys.executable, "-m", "pip", "install", "-U", "."])
    #     chdir(cwd)
   # download nltk stuff
    from os import getenv, path

    import nltk

    from rb.core.lang import Lang
    from rb.utils.downloader import download_wordnet

    nltk.download('punkt')
    nltk.download('wordnet')
    nltk.download('omw')

    nltk_path = getenv('NLTK_DATA')
    if nltk_path is None:
        nltk_path = path.expanduser('~/nltk_data/')
    download_wordnet(Lang.RO, nltk_path + "corpora/omw")


class PostDevelopmentCommand(develop):
    """Post-instalation for development? mode."""
    def run(self):
        develop.run(self)
        do_post_install_tasks()

class PostInstallCommand(install):
    """Post-instalation for install mode."""
    def run(self):
        install.run(self)
        do_post_install_tasks()
        
with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name='rbpy-rb',
    version='0.10.43',
    python_requires='>=3.6,<3.9',
    author='Woodcarver',
    author_email='batpepastrama@gmail.com',
    description='ReaderBench library written in python',
    long_description=long_description,
    long_description_content_type='text/markdown',
    url='https://git.readerbench.com/ReaderBench/Readerbench-python',
    packages=setuptools.find_packages(),
    include_package_data=True,
    install_requires=required,
    data_files=[("",["requirements.txt"])],
    cmdclass={
        'develop': PostDevelopmentCommand,
        'install': PostInstallCommand,
    },
    classifiers=[
        "Programming Language :: Python :: 3",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent"
    ],
)



import json
from typing import Dict, List

import xmltodict
from copy import deepcopy
from datetime import datetime

from rb.core.block import Block
from rb.core.cscl.contribution import Contribution
from rb.core.document import Document
from rb.core.lang import Lang
from rb.core.sentence import Sentence
from rb.core.text_element import TextElement
from rb.core.text_element_type import TextElementType
from rb.core.cscl.participant import Participant
from rb.utils.rblogger import Logger

logger = Logger.get_logger()

CONTRIBUTIONS_KEY = 'contributions'
ID_KEY = 'id'
PARENT_ID_KEY = 'parent_id'
TIMESTAMP_KEY = 'timestamp'
TEXT_KEY = 'text'
USER_KEY = 'user'
CONV_ID = 'conv_id'

TIMEFRAME = 30
DISTANCE = 20

class Conversation(TextElement):

    '''
        conversation_thread -> JSON object (crawled from Reddit) parsed as Dict
        contains a list with one object per contribution
        each contribution contains: user (anonymous id)
                                    id
                                    parent_id
                                    timestamp
                                    text
    '''
    def __init__(self, lang: Lang, conversation_thread: Dict,
                 container: TextElement = None,
                 depth: int = TextElementType.DOC.value,
                 apply_heuristics: bool = True,
                 start_date: datetime = None, end_date: datetime = None
                 ):

        TextElement.__init__(self, lang=lang, text="",
                             depth=depth, container=container)

        self.participant_map = dict()
        self.participants = []
        self.participant_contributions = dict()

        if conversation_thread:
            self.parse_contributions(conversation_thread, apply_heuristics=apply_heuristics, 
                                     start_date=start_date, end_date=end_date)
            self.scores = dict()
            self.init_scores()


    def time_heuristic(self, contributions: List[Dict]) -> List[Dict]:
        last_contribution = dict()
        group = dict()
        processed_contributions = [contributions[0]]

        for i, contribution in enumerate(contributions[1:]):
            index = contribution[ID_KEY]
            user = contribution[USER_KEY]
            timestamp = contribution[TIMESTAMP_KEY]

            if user == processed_contributions[-1][USER_KEY] and (timestamp - processed_contributions[-1][TIMESTAMP_KEY]).seconds <= TIMEFRAME:
                processed_contributions[-1][TEXT_KEY] += (' ' + contribution[TEXT_KEY])
                group[index] = processed_contributions[-1]

            else:
                group[index] = contribution

                processed_contributions.append(contribution)
                
                parent_id = int(contribution[PARENT_ID_KEY])
                if parent_id > 0:
                    parent = group[parent_id]

                    contribution[PARENT_ID_KEY] = parent[ID_KEY]

        return processed_contributions

    def distance_heuristic(self, contributions: List[Dict]) -> List[Dict]:
        last_contribution = dict()
        last_contribution_index = dict()
        group = dict()
        processed_contributions = []

        for i, contribution in enumerate(contributions):
            user = contribution[USER_KEY]

            if user in last_contribution_index and ((i - last_contribution_index[user]) <= DISTANCE):
                last = last_contribution[user]
                last[TEXT_KEY] += (' ' + contribution[TEXT_KEY])
                group[i] = last

            else:
                last_contribution_index[user] = int(contribution[ID_KEY])
                contribution[ID_KEY] = len(processed_contributions)
                group[i] = contribution

                processed_contributions.append(contribution)
                last_contribution[user] = contribution
            
                parent_id = int(contribution[PARENT_ID_KEY])
                if parent_id > 0:
                    parent = group[parent_id]

                    contribution[PARENT_ID_KEY] = parent[ID_KEY]

        return processed_contributions

    def parse_contributions(self, conversation_thread: Dict, apply_heuristics: bool = True, start_date: datetime = None, end_date: datetime = None):
        contribution_map = dict()
        users = set()

        full_text = ''

        # apply both heuristics before processing
        contributions = conversation_thread[CONTRIBUTIONS_KEY]
        self.conv_id = conversation_thread[CONV_ID]
        if apply_heuristics:
            contributions = self.time_heuristic(contributions)
            # contributions = self.distance_heuristic(contributions)

        self.participants = []

        for contribution in contributions:
            index = int(contribution[ID_KEY])
            participant_id = contribution[USER_KEY]
            users.add(participant_id)

            # parent index will be -1 in JSON for the first post
            parent_index = int(contribution[PARENT_ID_KEY])

            timestamp = contribution[TIMESTAMP_KEY]
            if start_date and end_date and (start_date > timestamp or timestamp >= end_date):
                continue
            text = contribution[TEXT_KEY].strip()
            if text[-1] not in {'.', '!', '?'}:
                text += "."
            
            parent_contribution = None

            if parent_index > 0 and parent_index in contribution_map:
                parent_contribution = contribution_map[parent_index]

            full_text += text + "\n"

            participant = None

            if participant_id not in self.participant_map:
                participant = Participant(participant_id=participant_id)

                self.participant_map[participant_id] = participant
                self.participants.append(participant)
            else:
                participant = self.participant_map[participant_id]

            if self.container is not None:
                if participant_id not in self.container.participant_map:
                    global_participant = Participant(participant_id=participant_id)
                    self.container.participant_map[participant_id] = global_participant		

            current_contribution = Contribution(self.lang, text, container=self,
                                                participant=participant,
                                                parent_contribution=parent_contribution,
                                                contribution_raw=contribution,
                                                index=index,
                                                timestamp=timestamp)

            self.components.append(current_contribution)
            contribution_map[index] = current_contribution

            if not (participant_id in self.participant_contributions):
                self.participant_contributions[participant_id] = []

            self.participant_contributions[participant_id].append(current_contribution)

        self.text = full_text

        sentences = self.parse_full_text(full_text)
        i = 0
        for contribution in self.components:
            left = len(contribution.text)
            while i < len(sentences) and len(sentences[i].text) <= left:
                contribution.add_sentence(sentences[i])
                left -= len(sentences[i].text)
                i += 1


    def parse_full_text(self, full_text: str) -> List[Sentence]:
        parsed_document = Block(self.lang, full_text)
        return parsed_document.get_sentences()

    def get_participants(self) -> List[Participant]:
        return self.participants

    def get_participant_contributions(self, participant_id: str) -> List[Contribution]:
        return self.participant_contributions[participant_id]

    def get_contributions(self) -> List[Contribution]:
        return self.components

    def init_scores(self):
        for a in self.participants:
            self.scores[a.get_id()] = dict()

            for b in self.participants:
                self.scores[a.get_id()][b.get_id()] = 0

    def get_score(self, a: str, b: str) -> float:
        return self.scores[a][b]

    def update_score(self, a: str, b: str, value: float):
        self.scores[a][b] += value

    def __str__(self):
        return NotImplemented

    def __eq__(self, other):
        if not isinstance(other, Conversation):
            return False
        return self.conv_id == other.conv_id and super(TextElement, self).__eq__(other)

    def __hash__(self):
        return hash((self.conv_id, super(TextElement, self).__hash__()))
    
    @staticmethod
    def create_participant_conversation(lang: Lang, p: Participant) -> "Conversation":
        conversation_thread = dict()
        contribution_list = []

        index = 0
        old_index = dict()
        conversation_thread[CONV_ID] = p.get_id()
        for c in p.eligible_contributions:
            contribution = deepcopy(c.get_raw_contribution())

            old_index[contribution[ID_KEY]] = index
            contribution[ID_KEY] = index

            if contribution[PARENT_ID_KEY] in old_index:
                contribution[PARENT_ID_KEY] = old_index[contribution[PARENT_ID_KEY]]
            else:
                contribution[PARENT_ID_KEY] = -1

            contribution_list.append(contribution)
            index += 1

        conversation_thread[CONTRIBUTIONS_KEY] = contribution_list

        conversation = Conversation(lang=lang, container=None,
                                    conversation_thread=conversation_thread,
                                    apply_heuristics=False)

        p.set_own_conversation(conversation)

    
    @staticmethod
    def load_from_xml(lang: Lang, filename: str) -> "Conversation":
        with open(filename, "rt") as f:
            my_dict=xmltodict.parse(f.read())
            contributions = [
                {
                    ID_KEY: int(utterance["@genid"]),
                    PARENT_ID_KEY: int(utterance["@ref"]),
                    TIMESTAMP_KEY: utterance["@time"],
                    USER_KEY: turn["@nickname"],
                    TEXT_KEY: utterance["#text"],
                }
                for turn in my_dict["corpus"]["Dialog"]["Body"]["Turn"]
                for utterance in (turn["Utterance"] if isinstance(turn["Utterance"], List) else [turn["Utterance"]])
            ]
            return Conversation(lang, {CONTRIBUTIONS_KEY: contributions})


from collections import Counter
import sys
import re

def compute_sentences(filepath):
	sentences = 0
	with open(filepath, 'r', encoding='utf-8') as in_file:
		for _ in in_file:
			sentences += 1

	return sentences


def compute_words(filepath):
	words = 0
	with open(filepath, 'r', encoding='utf-8') as in_file:
		for line in in_file:
			words += len(line.split(" "))

	return words


def compute_chars(filepath):
	chars = 0
	with open(filepath, 'r', encoding='utf-8') as in_file:
		for line in in_file:
			chars += len(line)
	return chars


def compute_chars_dict(filepath):
	global_counter = Counter()
	with open(filepath, 'r', encoding='utf-8') as in_file:
		for line in in_file:
			global_counter += Counter(line)	
	return global_counter


def compute_chars_diacritics(filepath):
	regexPattern = re.compile('[aăâiîsștț]')
	chars = 0
	with open(filepath, 'r', encoding='utf-8') as in_file:
		for line in in_file:
			listOfmatches = regexPattern.findall(line)
			chars += len(listOfmatches)
	return chars


def compute_chars_no_spaces(filepath):
	chars = 0
	with open(filepath, 'r', encoding='utf-8') as in_file:
		for line in in_file:
			sentence = line.replace(" ", "")
			sentence = sentence.replace("\n", "")
			chars += len(sentence)
	return chars


def compute_statistics(filepath):
	print("Statistics for", filepath.split("/")[-1])

	sent = compute_sentences(filepath)
	words = compute_words(filepath)
	chars = compute_chars(filepath)
	chars_no_space = compute_chars_no_spaces(filepath)
	chars_dict = compute_chars_dict(filepath)
	chars_diacritics = compute_chars_diacritics(filepath)

	print("Sentences =", format(sent, ",d"))
	print("Words =", format(words, ",d"))
	print("Total chars =", format(chars, ",d"))
	print("Chars w/o spaces =", format(chars_no_space, ",d"))
	print("Char with possible diacritics =", format(chars_diacritics, ",d"))
	print("Chars dict =", chars_dict)
	print("Unique chars =", len(chars_dict))
	print("--------------")


if __name__ == "__main__":

	compute_statistics("dataset/split/train.txt")
	compute_statistics("dataset/split/dev.txt")
	compute_statistics("dataset/split/test.txt")